//////////
// Discrete vector analysis functionality
//////////

/*
getScalarField
Generate an n-dimensional discrete hypercubic scalar field of the given size.
Arguments:
	dimension: The dimension of the field
	size: The "width" of the hypercube containing the field
	defaultValue: The value all scalars in the field (so it's always flat and don't contain undefined values)
*/
function getScalarField(dimension, size, defaultValue)
{
	dimension = (dimension || 2);
	size = (size || mapSize + 1); // Default for heightmaps
	defaultValue = (defaultValue || 0);
	var scalarField = defaultValue;
	for (var d = 0; d < dimension; d++)
	{
		var newScalarField = [];
		for (var s = 0; s < size; s++)
			newScalarField.push(scalarField)
		scalarField = deepcopy(newScalarField);
	}
	return scalarField;
}

/*
getVectorField
Generate an n-dimensional discrete hypercubic vetor field of the given size.
Arguments:
	dimension: int, The dimension of the field
	size: int, The "width" of the hypercube containing the field (So float coordinates are not supported!)
	defaultValue: float, The value all vectors components in the field (So div and curl of this vector field is always 0 and vectors are colinear to (1, 1, 1, ...) or the 0-vector)
*/
function getVectorField(dimension, size, defaultValue)
{
	dimension = (dimension || 2);
	size = (size || mapSize + 1); // Default for heightmaps
	defaultValue = (defaultValue || 0);
	var vector = [];
	for (var d = 0; d < dimension; d++)
		vector.push(defaultValue);
	var vectorField = vector;
	for (var d = 0; d < dimension; d++)
	{
		var newVectorField = [];
		for (var s = 0; s < size; s++)
			newVectorField.push(vectorField)
		vectorField = deepcopy(newVectorField);
	}
	return vectorField;
}

/*
getDimensionOfField
Takes an discrete  hypercubic scalar or vector field and returns it's dimension.
*/
function getDimensionOfField(field)
{
	var nextDepthObject = deepcopy(field);
	var size = field.length;
	var dimension = 0;
	while (typeof(nextDepthObject) == "object")
	{
		if (nextDepthObject.length == size)
		{
			dimension++;
			nextDepthObject = nextDepthObject[0];
		}
		else
			break;
	}
	return dimension;
}

function getAllCoordinatesInHypercube(dimension, size)
{
	var coordinates = [];
	for (var i = 0; i < Math.pow(size, dimension); i++)
	{
		var vector = [];
		var sum = 0
		for (var d = 0; d < dimension; d++)
		{
			vector.push(((i - sum) % (Math.pow(size, d + 1))) / Math.pow(size, d));
			sum += vector[d]*Math.pow(size, d);
		}
		coordinates.push(vector);
	}
	return coordinates;
}

function getVectorLength(vector)
{
	var squareSum = 0;
	for (var d = 0; d < vector.length; d++)
		squareSum += Math.pow(vector[d], 2);
	return Math.pow(squareSum, 1/2);
}

function getFieldValueByCoordinate(field, vector)
{
	var dimension = vector.length;
	var size = field.length;
	var value = deepcopy(field);
	for (var d = 0; d < dimension; d++)
	{
		value = value[vector[d]];
	}
	return value;
}

function setFieldValueByCoordinate(field, vector, value)
{
	var coordinate = vector.shift();
	if (vector.length == 0)
		field[coordinate] = value;
	else
		setFieldValueByCoordinate(field[coordinate], vector, value);
}

function getNormalizedField(field, scaleToLength, isVectorField)
{
	scaleToLength = (scaleToLength || 1);
	var dimension = getDimensionOfField(field);
	var size = field.length;
	var coordinates = getAllCoordinatesInHypercube(dimension, size);
	isVectorField = (isVectorField || typeof(getFieldValueByCoordinate(field, coordinates[0])) == "object");
	var maxLength = 0;
	for (var i = 0; i < coordinates.length; i++)
	{
		if (isVectorField)
			var length = getVectorLength(getFieldValueByCoordinate(field, coordinates[i]));
		else
			var length = abs(getFieldValueByCoordinate(field, coordinates[i]));
		if (length > maxLength)
			maxLength = length;
	}
	var newField = deepcopy(field);
	for (var i = 0; i < coordinates.length; i++)
	{
		if (isVectorField)
		{
			var newValue = [];
			for (var d = 0; d < dimension; d++)
				newValue.push(getFieldValueByCoordinate(field, coordinates[i])[d] / maxLength * scaleToLength);
		}
		else
			var newValue = getFieldValueByCoordinate(field, coordinates[i]) / maxLength * scaleToLength;
		setFieldValueByCoordinate(newField, coordinates[i], newValue);
	}
	return newField
}

/*
getGrad
Returns the gradient of a given discrete hypercuibic scalar field, a discrete hypercubic vector field (gradient field). Dimension might depend on method: scalar fields width or the same - 1.
Problems:
	The calculated vectorfield would be of size sizeOfScalarField - 1. To avoid this the left is wrapped to the right and the bottom to the top.
	Using 2 appending fields results in a shift of shiftVector = 1/2*(1, 1, 1, ... , 1), abs(shiftVector) = 1/2*sqareroot(dimension) between the given scalar and the resulting vector field.
	Using the 2 fiels appending the given tile avoids the shift but results in 2 areas independent of each other shifted towards each other by 1. I don't know if that is bad.
	It might be a good idea to include the diagonal appending tiles for calculation to couple those two independent areas...
*/
function getGrad(scalarField, dimension, size)
{
	dimension = (dimension || getDimensionOfField(scalarField));
	size = (size || scalarField.length);
	var vectorField = getVectorField(dimension, size);
	var coordinates = getAllCoordinatesInHypercube(dimension, size);
	for (var i = 0; i < coordinates.length; i++)
	{
		var vector = [];
		for (var d = 0; d < dimension; d++)
		{
			var addCoordinate = deepcopy(coordinates[i]);
			addCoordinate[d] = (addCoordinate[d] + 1) % size;
			var subCoordinate = deepcopy(coordinates[i]);
			subCoordinate[d] = (subCoordinate[d] - 1 + size) % size;
			vector.push(getFieldValueByCoordinate(scalarField, addCoordinate) - getFieldValueByCoordinate(scalarField, subCoordinate));
		}
		setFieldValueByCoordinate(vectorField, coordinates[i], vector);
	}
	return vectorField;
}

// Same problems like in getGrad so all dimensions wrapped and the actual coordinates value ignored
function getDiv(vectorField, dimension, size)
{
	dimension = (dimension || getDimensionOfField(vectorField));
	size = (size || vectorField.length);
	var scalarField = getVectorField(dimension, size);
	var coordinates = getAllCoordinatesInHypercube(dimension, size);
	for (var i = 0; i < coordinates.length; i++)
	{
		var value = 0;
		for (var d = 0; d < dimension; d++)
		{
			var addCoordinate = deepcopy(coordinates[i]);
			addCoordinate[d] = (addCoordinate[d] - 1 + size) % size;
			value += getFieldValueByCoordinate(vectorField, addCoordinate)[d];
			var subCoordinate = deepcopy(coordinates[i]);
			subCoordinate[d] = (subCoordinate[d] + 1) % size;
			value -= getFieldValueByCoordinate(vectorField, subCoordinate)[d];
		}
		setFieldValueByCoordinate(scalarField, coordinates[i], value);
	}
	return scalarField;
}
